package it.condarelli.writer.model.world;

import it.condarelli.writer.model.world.util.WorldAdapterFactory;
import it.condarelli.writer.model.world.util.WorldResourceFactoryImpl;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.ArrayList;
import java.util.EventObject;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.eclipse.emf.common.command.BasicCommandStack;
import org.eclipse.emf.common.command.Command;
import org.eclipse.emf.common.command.CommandStack;
import org.eclipse.emf.common.command.CommandStackListener;
import org.eclipse.emf.common.util.Diagnostic;
import org.eclipse.emf.common.util.URI;
import org.eclipse.emf.ecore.EClassifier;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.impl.EReferenceImpl;
import org.eclipse.emf.ecore.resource.Resource;
import org.eclipse.emf.ecore.resource.ResourceSet;
import org.eclipse.emf.ecore.resource.impl.ResourceSetImpl;
import org.eclipse.emf.ecore.util.Diagnostician;
import org.eclipse.emf.ecore.xmi.XMLResource;
import org.eclipse.emf.edit.domain.AdapterFactoryEditingDomain;
import org.eclipse.emf.edit.domain.EditingDomain;

public class Model {
	protected ResourceSet	resourceSet;
	protected Resource		resource;

	protected Model() {
		// Create a resource set to hold the resources.
		resourceSet = new ResourceSetImpl();

		// Register the appropriate resource factory to handle all file
		// extensions.
		resourceSet.getResourceFactoryRegistry().getExtensionToFactoryMap().put(World.DEFAULT_EXTENSION, new WorldResourceFactoryImpl());

		// Register the package to ensure it is available during loading.
		resourceSet.getPackageRegistry().put(WorldPackage.eNS_URI, WorldPackage.eINSTANCE);

		// ignore leftovers
		resourceSet.getLoadOptions().put(XMLResource.OPTION_RECORD_UNKNOWN_FEATURE, Boolean.TRUE);

	}

	public Model(URI uri) {
		this();
		load(uri);
	}

	public void load(String fname) {
		if (fname == null) {
			Resource resource = resourceSet.createResource(URI.createURI("http:///new." + World.DEFAULT_EXTENSION));
			World root = WorldFactory.eINSTANCE.createWorld();
			resource.getContents().add(root);
			root.setModel(this);
		} else {
			File file = new File(fname);
			URI uri = file.isFile() ? URI.createFileURI(file.getAbsolutePath()) : URI.createURI(fname);
			load(uri);
		}
	}

	public void load(InputStream is) {
		resource = resourceSet.createResource(URI.createURI("http:///new." + World.DEFAULT_EXTENSION));
		Map<String, String> opt = new HashMap<String, String>();
		opt.put(XMLResource.OPTION_ENCODING, "UTF-8");
		try {
			resource.load(is, opt);
			check();
			getWorld().setModel(this);
		} catch (IOException e) {
			e.printStackTrace();
		}
	}

	protected void load(URI uri) {
		try {
			// Demand load resource for this file.
			resource = resourceSet.getResource(uri, true);
			// System.out.println("Loaded " + uri);
		} catch (RuntimeException exception) {
			System.err.println("Problem loading " + uri);
			exception.printStackTrace();
		}
		check();
		getWorld().setModel(this);
	}

	public Resource getResource() {
		return resource;
	}

	public void save() throws IOException {
		if (resource.isModified()) {
			Map<String, String> opt = new HashMap<String, String>();
			opt.put(XMLResource.OPTION_ENCODING, "UTF-8");
			resource.save(opt);
			lastCommand = iedp.getCommandStack().getMostRecentCommand();
		}
	}

	public void saveAs(String fname) throws IOException {
		File file = new File(fname);
		URI uri = URI.createFileURI(file.getAbsolutePath());
		resource.setURI(uri);
		resource.setModified(true);
		save();
	}

	public void save(OutputStream os) {
		try {
			Map<String, String> opt = new HashMap<String, String>();
			opt.put(XMLResource.OPTION_ENCODING, "UTF-8");
			resource.save(os, opt);
		} catch (IOException e) {
			e.printStackTrace();
		}
	}

	EditingDomain	iedp			= null;
	Command			lastCommand	= null;

	public EditingDomain getEditingDomain() {
		if (iedp == null) {
			final CommandStack cs = new BasicCommandStack();
			iedp = new AdapterFactoryEditingDomain(new WorldAdapterFactory(), cs);
			// System.out.println("WorldEditingDomain.WorldEditingDomain():");
			cs.addCommandStackListener(new CommandStackListener() {
				@Override
				public void commandStackChanged(EventObject event) {
					boolean dirty = (lastCommand == null) ? cs.canUndo() : (lastCommand == cs.getMostRecentCommand());
					if (dirty != resource.isModified()) {
						resource.setModified(dirty);
						notifyDirty(dirty);
					}
				}
			});
		}
		return iedp;
	}

	public interface DirtyListener {
		public void setDirty(boolean dirty);
	}

	private List<DirtyListener>	dirtyListeners	= new ArrayList<DirtyListener>();

	public void addDirtyListener(DirtyListener listener) {
		dirtyListeners.add(listener);
		listener.setDirty(resource.isModified());
	}

	public void removeDirtyListener(DirtyListener listener) {
		dirtyListeners.remove(listener);
	}

	protected void notifyDirty(boolean dirty) {
		for (DirtyListener dl : dirtyListeners) {
			dl.setDirty(dirty);
		}
	}

	public boolean isDirty() {
		return resource.isModified();
	}

	public World getWorld() {
		for (EObject eObject : resource.getContents()) {
			if (eObject instanceof World) {
				return (World) eObject;
			}
		}
		return null;
	}

	protected void check() {
		if (resource instanceof XMLResource) {
			XMLResource xr = (XMLResource) resource;
			// for (EObject eo : xr.getEObjectToExtensionMap().keySet())
			// System.out.println(eo);
			xr.getEObjectToExtensionMap().clear();
		}
		for (EObject eObject : resource.getContents()) {
			Diagnostic diagnostic = Diagnostician.INSTANCE.validate(eObject);
			if (diagnostic.getSeverity() != Diagnostic.OK) {
				printDiagnostic(diagnostic, "");
			}
			// walkModel(eObject, "   ");
		}

	}

	protected static void printDiagnostic(Diagnostic diagnostic, String indent) {
		System.out.print(indent);
		System.out.println(diagnostic.getMessage());
		System.out.print(indent);
		System.out.print("--");
		int code = diagnostic.getCode();
		EObject eo = (EObject) diagnostic.getData().get(0);
		System.out.printf("code:%2d sender:%s ", code, eo.getClass());
		switch (code) {
		case 0:
			// do nothing
			break;
		case 1: // missing required feature
			EReferenceImpl eri = (EReferenceImpl) diagnostic.getData().get(1);
			EClassifier ec = eri.getEType();
			// EPackage ep = ec.getEPackage();
			Class<?> chc = ec.getInstanceClass();
			// EClass chec = ec.eClass();
			System.out.printf("child class: %s ", chc);
			// EObject cho = ep.getEFactoryInstance().create(chec);
			if (chc.equals(Timeline.class)) {
				EObject cho = WorldFactory.eINSTANCE.createTimeline();
				eo.eSet(eri, cho);
				System.out.print("done.");
				break;
			}
			if (chc.equals(Actor.class)) {
				if (eo instanceof Scene) {
					Scene sc = (Scene) eo;
					Actor a = sc.getActors().get(0);
					sc.getActor().add(a);
					System.out.print("done.");
					break;
				}
			}
			System.out.print("unknown.");
			break;
		default:
			System.out.print("unhandled code");
		}
		System.out.println();
		for (Diagnostic child : diagnostic.getChildren()) {
			printDiagnostic(child, indent + "  ");
		}
	}

	protected static void walkModel(EObject eo, String indent) {
		System.out.print(indent);
		System.out.println(eo.eClass().getInstanceClassName());
		for (EObject child : eo.eContents()) {
			walkModel(child, indent + "   ");
		}
	}
}
