package it.condarelli.writer.model.world;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.FilenameFilter;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.Calendar;
import java.util.Collections;
import java.util.Comparator;
import java.util.Date;
import java.util.GregorianCalendar;
import java.util.LinkedList;
import java.util.List;

import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Status;
import org.eclipse.core.runtime.jobs.Job;
import org.eclipse.emf.common.notify.Adapter;
import org.eclipse.emf.common.notify.Notification;
import org.eclipse.emf.common.notify.impl.AdapterImpl;
import org.eclipse.emf.common.util.URI;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.swt.widgets.Display;

public class SafeModel extends Model {
    private final class ChangeMonitor extends AdapterImpl {
        boolean updating = false;

        @Override
        public void notifyChanged(Notification msg) {
            super.notifyChanged(msg);
            if (updating)
                return;
            System.out.println("notifyChanged: " + msg);
            synchronized (saveDate) {
                changeDate = new Date();
            }
            if (saveJob == null) {
                saveJob = new Job("saveJob") {
                    @Override
                    protected IStatus run(IProgressMonitor monitor) {
                        Date now = new Date();
                        System.out.println("saveJob():");
                        if (tmp == null)
                            try {
                                newTmp();
                            } catch (FileNotFoundException e) {
                                String s = "saveJob(): unable to get tmp file: '" + e.getMessage() + "'";
                                System.err.println(s);
                                return new Status(Status.ERROR, "it.condarelli.writer.model", s);
                            }
                        updating = true;
                        saveTmp();
                        updating = false;
                        System.out.println(tmp.getAbsolutePath());
                        synchronized (saveDate) {
                            if (changeDate.before(now)) {
                                changeDate = null;
                            } else {
                                if (saveTmpDelay > 0)
                                    saveJob.schedule(saveTmpDelay);
                            }
                        }
                        saveDate = now;
                        return Status.OK_STATUS;
                    }
                };
            } else {
                saveJob.cancel();
            }
            if (saveTmpDelay >= 0)
                saveJob.schedule(saveTmpDelay);
        }
    }

    private File    tmp;
    private Date    changeDate;
    private Date    saveDate;
    private Job     saveJob;

    private int     saveTmpDelay = -1;
    // private int saveOldInterval = -1;
    // private int maxOldVersions = -1;
    private Adapter mta;

    protected File computeTmp() throws FileNotFoundException {
        URI uTmp = getResource().getURI().trimFileExtension().appendFileExtension("wtmp");
        tmp = checkWritable(uTmp);
        return tmp;
    }

    protected File newTmp() throws FileNotFoundException {
        if (tmp == null)
            computeTmp();
        safeRename();
        return tmp;
    }

    protected File computeOld() throws FileNotFoundException {
        URI uTmp = getResource().getURI();
        String sOld = uTmp.toString();
        Calendar now = new GregorianCalendar();
        String tag = String.format("-%04d%02d%02d", now.get(Calendar.YEAR), now.get(Calendar.MONTH) + 1, now.get(Calendar.DAY_OF_MONTH));
        sOld = sOld.replaceFirst("\\.([^./])+$", tag + ".$1");

        File old = checkWritable(uTmp);
        return old;
    }

    protected File checkWritable(URI uri) throws FileNotFoundException {
        File file;
        if (uri.isFile()) {
            file = new File(uri.toFileString());
        } else {
            String fn = uri.lastSegment();
            File td = new File(System.getProperty("java.library.path"));
            file = new File(td, fn);
        }
        if (!file.getParentFile().exists()) {
            String s = String.format("%s.computeTmp(): error; parent directory for '%s' does not exist", this.getClass().getName(), file.getPath());
            System.err.println(s);
            file = null;
            throw new FileNotFoundException(s);
        }
        if (file.exists() && !file.canWrite()) {
            String s = String.format("%s.computeTmp(): error; unable to write to '%s'", this.getClass().getName(), file.getPath());
            System.err.println(s);
            file = null;
            throw new FileNotFoundException(s);
        }
        return file;

    }

    public void saveTmp() {
        File t = null;
        if (tmp.exists()) {
            // already saved; make sure we do not lose all!
            File old = new File(tmp.toURI());
            try {
                t = File.createTempFile("SafeModel", null);
                if (t.exists())
                    t.delete();
                if (old.renameTo(t)) {
                    tmp = old;
                } else {
                    System.err.println("unable to rename '" + tmp.toString() + " to '" + t.toString() + "'");
                    t = null;
                }
            } catch (IOException e) {
                // TODO Auto-generated catch block
                e.printStackTrace();
            }
        }
        try {
            OutputStream os = new FileOutputStream(tmp);
            save(os);
            try {
                os.close();
            } catch (IOException e) {
                System.err.println("unable to close output stream: " + e.getMessage());
            }
            if (t != null) {
                t.delete();
                t = null;
            }
        } catch (FileNotFoundException e) {
            System.err.println("TmpFile.save(): error: " + e.getMessage());
            if (t != null) {
                if (tmp.exists())
                    tmp.delete();
                if (!t.renameTo(tmp)) {
                    System.err.println("unable to rename '" + t.toString() + " to '" + tmp.toString() + "'");
                } else {
                    t = null;
                }
            }
        }
        if (t != null) {
            final String p = t.getAbsolutePath();
            Display.getDefault().syncExec(new Runnable() {
                @Override
                public void run() {
                    MessageDialog.openError(Display.getDefault().getActiveShell(), "Something went seriously wrong while saving tmp file", "file '" + p
                            + "' should be a good copy of model");
                }
            });
        }
    }

    public void save() throws IOException {
        boolean stale = true;
        if ((saveDate != null) && ((changeDate == null) || saveDate.after(changeDate))) {
            stale = false;
            // saved in tmp; just rename tmp to real filename.
            System.out.println("doSave(): renaming");
            URI uri = getResource().getURI();
            if ((tmp != null) && tmp.exists() && (tmp.length() > 0)) {
                File res = new File(uri.toFileString());
                if (res.exists())
                    res.delete();
                if (tmp.renameTo(res)) {
                    tmp = null;
                    return;
                }
            }
        }

        // Either rename failed or tmp is stale; try a full save.
        System.out.println("doSave(): full save");
        super.save();

        if (tmp != null) {
            if (!stale)
                MessageDialog.openWarning(Display.getCurrent().getActiveShell(), "Non stale tmp file still present", "Check '" + tmp.getAbsolutePath()
                        + "'\nNOTE: it wil be removed when You hit [ Ok ].");
            tmp.delete();
            tmp = null;
        }
    }

    protected List<File> getTemps(File parent, final String pattern) {
        List<File> fl = new LinkedList<File>();
        if ((parent != null) && parent.isDirectory()) {
            for (File f : parent.listFiles(new FilenameFilter() {
                @Override
                public boolean accept(File dir, String name) {
                    return (name.matches(pattern));
                }
            })) {
                fl.add(f);
            }
            if (!fl.isEmpty()) {
                Collections.sort(fl, new Comparator<File>() {
                    @Override
                    public int compare(File o1, File o2) {
                        return (int) (o1.lastModified() - o2.lastModified());
                    }
                });
            }
        }
        return fl;
    }

    protected boolean checkRestorable(URI uri) {
        URI uTmp = uri.trimFileExtension().appendFileExtension("wtmp");
        final String n = "\\Q" + uTmp.lastSegment() + "\\E.*";
        File parent = (new File(uTmp.toFileString())).getParentFile();

        List<File> fl = getTemps(parent, n);
        if (fl.isEmpty()) {
            fl = getTemps(new File(System.getProperty("java.io.tmpdir")), n);
        }

        File old = null;
        for (File f : fl) {
            if ((old != null) && (old.lastModified() > f.lastModified())) {
                // do nothing
            } else {
                old = f;
            }
        }
        if (old != null) {
            File f = new File(uri.toFileString());
            if (!f.exists() || (f.lastModified() < old.lastModified()))
                if (MessageDialog.openQuestion(Display.getDefault().getActiveShell(), "Recovery", "Old leftover found; Initiate recovery)?")) {
                    tmp = old;
                    return true;
                }
            old.deleteOnExit();
        }

        return false;
    }

    public SafeModel(URI uri) {
        super();
        if (checkRestorable(uri)) {
            try {
                InputStream is = new FileInputStream(tmp);
                load(is);
                saveDate = new Date(tmp.lastModified());
                resource.setURI(uri);
                monitor();
                return;
            } catch (FileNotFoundException e) {
            }
        }
        load(uri);
        if (uri.isFile()) {
            File f = new File(uri.toFileString());
            saveDate = new Date(f.lastModified());
        }
        monitor();
    }

    protected void monitor() {
        if (saveTmpDelay >= 0) {
            getResource().setTrackingModification(true);
            mta = new ChangeMonitor();
            getResource().eAdapters().add(mta);
        } else {
            if (mta != null) {
                getResource().eAdapters().remove(mta);
                mta = null;
            }
        }
    }

    public int getSaveTmpDelay() {
        return saveTmpDelay;
    }

    public void setSaveTmpDelay(int saveTmpDelay) {
        this.saveTmpDelay = saveTmpDelay;
        monitor();
    }

    protected void safeRename() throws FileNotFoundException {
        safeRename(tmp);
    }

    protected void safeRename(File t) throws FileNotFoundException {
        if (t.exists()) {
            String n = t.getAbsolutePath() + "~";
            File f = new File(n);
            safeRename(f);
            if (f.exists())
                throw new FileNotFoundException("unable to free '" + f.getAbsolutePath() + "'");
            File r = new File(t.toURI());
            r.renameTo(f);
            if (t.exists())
                throw new FileNotFoundException("file '" + f.getAbsolutePath() + "' still exists after rename");
        }
    }

    public boolean hasTmp() {
        return (tmp != null) && tmp.exists() && (tmp.length() > 0);
    }
}
